import torchvision.transforms as T


def get_image_transform(
    image_size: int,
    center_crop: bool = False,
    interpolation: T.InterpolationMode = T.InterpolationMode.BILINEAR,  # We used bilinear during training
):
    if center_crop:
        crop = [T.Resize(image_size, interpolation=interpolation), T.CenterCrop(image_size)]
    else:
        # "Squash": most versatile
        crop = [T.Resize((image_size, image_size), interpolation=interpolation)]

    return T.Compose(
        crop
        + [
            T.Lambda(lambda x: x.convert("RGB")),
            T.ToTensor(),
            T.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5], inplace=True),
        ]
    )



"""
Encoder Class for DUNE

DUNE uses the same implementation as DINOv2, where the only difference is the loaded weights.
"""

from typing import List, Optional, Union

import torch
import torch.nn.functional as F
from torch import nn

from uniception.models.encoders.base import UniCeptionViTEncoderBase, ViTEncoderInput, ViTEncoderOutput
from uniception.models.utils.intermediate_feature_return import IntermediateFeatureReturner


class DUNEEncoder(UniCeptionViTEncoderBase):
    "UniCeption DUNE Encoder"

    def __init__(
        self,
        name: str,
        pretrained_checkpoint_path: str,
        data_norm_type: str = "dune",
        patch_size: int = 14,
        vit_size: str = "base",
        pe_image_size: int = 448,
        torch_hub_force_reload: bool = False,
        gradient_checkpointing: bool = False,
        keep_first_n_layers: Optional[int] = None,
        use_pytorch_sdpa=True,
        disable_torch_compile_for_pe=False,
        *args,
        **kwargs,
    ):
        """
        DUNE Encoder for extracting spatial features from images.
        DUNE uses the same implementation as DINOv2 with registers, with custom pretrained weights.

        Args:
            name (str): Name of the encoder.
            pretrained_checkpoint_path (str): Path to the pretrained DUNE checkpoint.
            data_norm_type (str): Image normalization type. Default: "dune"
            patch_size (int): Patch size for the encoder. Default: 14
            vit_size (str): Size variant of the ViT model. Default: "base"
            pe_image_size (int): Image size for position encoding. Default: 448
            torch_hub_force_reload (bool): Whether to force reload the model from torch hub. Default: False
            gradient_checkpointing (bool): Whether to use gradient checkpointing to save GPU memory during backward call. Default: False
            keep_first_n_layers (Optional[int]): If specified, only the first n layers of the model will be kept. Default: None
            use_pytorch_sdpa (bool): Whether to use PyTorch native SDPA for attention layers. Default: True
            disable_torch_compile_for_pe (bool): Whether to disable torch compile for PE interpolation. Default: False
        """
        size = vit_size
        with_registers = True  # all DUNE encoder have registers

        # Init the base class
        name = name if not with_registers else f"{name}_reg"
        super().__init__(
            name=name,
            data_norm_type=data_norm_type,
            patch_size=patch_size,
            gradient_checkpointing=gradient_checkpointing,
            *args,
            **kwargs,
        )

        # Init the DINOv2 Encoder specific attributes
        self.version = size
        self.with_registers = with_registers
        self.enc_embed_dim = {"small": 384, "base": 768, "large": 1024, "giant": 1536}[self.version]

        # Define DINOv2 model factory
        DINO_MODELS = {
            # No registers
            False: {
                "small": "dinov2_vits14",
                "base": "dinov2_vitb14",
                "large": "dinov2_vitl14",
                "giant": "dinov2_vitg14",
            },
            # With registers
            True: {
                "small": "dinov2_vits14_reg",
                "base": "dinov2_vitb14_reg",
                "large": "dinov2_vitl14_reg",
                "giant": "dinov2_vitg14_reg",
            },
        }

        # Load the pretrained DINOv2 model from torch hub
        print(f"Loading pretrained {DINO_MODELS[self.with_registers][self.version]} from torch hub")
        try:  # Requires internet access
            self.model = torch.hub.load(
                "facebookresearch/dinov2",
                DINO_MODELS[self.with_registers][self.version],
                force_reload=torch_hub_force_reload,
            )
        except:  # Load from cache
            self.model = torch.hub.load(
                "facebookresearch/dinov2",
                DINO_MODELS[self.with_registers][self.version],
            )

        del (
            self.model.mask_token
        )  # This parameter is unused in producing patch features, and will lead to unused parameters

        # Patch DINOv2 position encoding table to have the correct shape
        num_patches = (pe_image_size // 14) ** 2
        self.model.pos_embed = torch.nn.Parameter(
            torch.zeros(1, num_patches + self.model.num_tokens, self.enc_embed_dim)
        )

        # disable position interpolation at PE interpolation to enable torch compile
        # when training with multiple image shapes.
        if disable_torch_compile_for_pe:
            self.model.interpolate_pos_encoding = torch.compiler.disable(self.model.interpolate_pos_encoding)

        # Keep only the first n layers of the model if keep_first_n_layers is specified
        if keep_first_n_layers is not None:
            self.model.blocks = nn.ModuleList(self.model.blocks[:keep_first_n_layers])

        # Use Native Torch SDPA for attention layers if specified (instead of DINOv2's XFormers)
        if use_pytorch_sdpa:
            self.enable_pytorch_native_sdpa()

        # Wrap the transformer blocks with support for gradient checkpointing if required
        if self.gradient_checkpointing:
            for i in range(len(self.model.blocks)):
                self.model.blocks[i] = self.wrap_module_with_gradient_checkpointing(self.model.blocks[i])

        # Load the custom pretrained checkpoint if provided
        if pretrained_checkpoint_path:
            print(f"Loading DUNE pretrained checkpoint from {pretrained_checkpoint_path}")
            ckpt = torch.load(pretrained_checkpoint_path, weights_only=False)
            # Extract and remap encoder weights from DUNE checkpoint
            encoder_state_dict = self._extract_and_remap_encoder_weights(ckpt["model"])
            print(self.load_state_dict(encoder_state_dict, strict=False))

    def enable_pytorch_native_sdpa(self):
        "Enable PyTorch native SDPA for attention layers"
        for i in range(len(self.model.blocks)):
            self.model.blocks[i].attn = self.wrap_dinov2_attention_with_sdpa(self.model.blocks[i].attn)

    def wrap_dinov2_attention_with_sdpa(self, module: nn.Module):
        "Wrap DINOv2 attention module with PyTorch native SDPA"
        assert torch.__version__ >= "2.0", "SDPA requires PyTorch 2.0 or later"

        class _AttentionWrapper(module.__class__):
            "SDPA Attention Wrapper Class"

            def forward(self, x: torch.Tensor, attn_bias=None) -> torch.Tensor:
                B, N, C = x.shape
                qkv = (
                    self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
                )  # (3, B, H, N, C // H)

                q, k, v = torch.unbind(qkv, 0)  # (B, H, N, C // H)

                x = F.scaled_dot_product_attention(q, k, v, attn_bias)
                x = x.permute(0, 2, 1, 3).reshape(B, N, C)

                x = self.proj(x)
                x = self.proj_drop(x)
                return x

        module.__class__ = _AttentionWrapper
        return module

    def _extract_and_remap_encoder_weights(self, checkpoint):
        """
        Extract encoder weights from DUNE checkpoint and remap keys to match DINOv2 model structure.

        DUNE checkpoint structure:
        - Keys are prefixed with "encoder." instead of "model."
        - Blocks are structured as "encoder.blocks.0.0." instead of "model.blocks.0."
        - Checkpoint also contains projectors and teacher_norms that we don't need

        Args:
            checkpoint: Loaded DUNE checkpoint dictionary

        Returns:
            dict: Remapped state dict with "model." prefix and correct block structure
        """
        encoder_state_dict = {}

        for key, value in checkpoint.items():
            # Only process encoder keys, skip projectors and teacher_norms
            if not key.startswith("encoder."):
                continue

            # Remove "encoder." prefix
            new_key = key.replace("encoder.", "", 1)

            # Fix blocks structure: "blocks.0.0." -> "blocks.0."
            # The checkpoint has blocks structured as blocks.0.0, blocks.0.1, etc.
            # We need to flatten this to blocks.0, blocks.1, etc.
            if "blocks." in new_key:
                parts = new_key.split(".")
                if len(parts) >= 3 and parts[0] == "blocks":
                    # Extract block indices (e.g., "0" and "0" from "blocks.0.0")
                    major_idx = parts[1]
                    minor_idx = parts[2]
                    # Combine indices: blocks.0.0 means the first block in the checkpoint
                    # We'll just use the minor index as the actual block number
                    # since blocks are numbered sequentially in the checkpoint
                    parts[1] = minor_idx
                    parts.pop(2)  # Remove the major index
                    new_key = ".".join(parts)

            # Add "model." prefix to match our model structure
            new_key = "model." + new_key
            encoder_state_dict[new_key] = value

        # Remove the mask token from the state dict
        # This parameter is unused in producing patch features, and will lead to unused parameters
        del encoder_state_dict["model.mask_token"]

        return encoder_state_dict

    def forward(self, encoder_input: ViTEncoderInput) -> ViTEncoderOutput:
        """
        DUNE Encoder Forward Pass

        Args:
            encoder_input (ViTEncoderInput): Input data for the encoder. Input data must contain image normalization type and normalized image tensor.

        Returns:
            ViTEncoderOutput: Output data from the encoder.
        """
        # Check image normalization type
        self._check_data_normalization_type(encoder_input.data_norm_type)

        # Check the dtype and shape of the input image
        assert isinstance(encoder_input.image, torch.Tensor), "Input must be a torch.Tensor"
        assert encoder_input.image.ndim == 4, "Input must be of shape (B, C, H, W)"
        batch_size, channels, height, width = encoder_input.image.shape
        assert channels == 3, "Input must have 3 channels"
        assert (
            height % self.patch_size == 0 and width % self.patch_size == 0
        ), f"Input shape must be divisible by patch size: {self.patch_size}"

        # Extract the features from the DINOv2 model
        features = self.model.forward_features(encoder_input.image)["x_norm_patchtokens"]

        # Resize the features to the expected shape
        # (B x Num_patches x Embed_dim) -> (B x Embed_dim x H / Patch_Size x W / Patch_Size)
        features = features.permute(0, 2, 1)
        features = features.reshape(
            -1, self.enc_embed_dim, height // self.patch_size, width // self.patch_size
        ).contiguous()

        return ViTEncoderOutput(features=features)


class DUNEIntermediateFeatureReturner(DUNEEncoder, IntermediateFeatureReturner):
    "Intermediate Feature Returner for UniCeption DUNE Encoder"

    def __init__(
        self,
        name: str,
        pretrained_checkpoint_path: str,
        data_norm_type: str = "dune",
        patch_size: int = 14,
        vit_size: str = "base",
        pe_image_size: int = 448,
        torch_hub_force_reload: bool = False,
        gradient_checkpointing: bool = False,
        keep_first_n_layers: Optional[int] = None,
        use_pytorch_sdpa=True,
        disable_torch_compile_for_pe=False,
        indices: Optional[Union[int, List[int]]] = 1,
        norm_intermediate: bool = True,
        *args,
        **kwargs,
    ):
        """
        DUNE Encoder for extracting spatial features from images with intermediate feature return.
        DUNE uses the same implementation as DINOv2 with registers, with custom pretrained weights.

        Args:
            name (str): Name of the encoder.
            pretrained_checkpoint_path (str): Path to the pretrained DUNE checkpoint.
            data_norm_type (str): Image normalization type. Default: "dune"
            patch_size (int): Patch size for the encoder. Default: 14
            vit_size (str): Size variant of the ViT model. Default: "base"
            pe_image_size (int): Image size for position encoding. Default: 448
            torch_hub_force_reload (bool): Whether to force reload the model from torch hub. Default: False
            gradient_checkpointing (bool): Whether to use gradient checkpointing to save GPU memory during backward call. Default: False
            keep_first_n_layers (Optional[int]): If specified, only the first n layers of the model will be kept. Default: None
            use_pytorch_sdpa (bool): Whether to use PyTorch native SDPA for attention layers. Default: True
            disable_torch_compile_for_pe (bool): Whether to disable torch compile for PE interpolation. Default: False
            indices (Optional[Union[int, List[int]]], optional): Indices of the layers to return. Defaults to 1. Options:
            - int: Return the last n layers.
            - List[int]: Return the intermediate layers at the specified indices.
            norm_intermediate (bool, optional): Whether to normalize the intermediate features. Defaults to True.
        """
        DUNEEncoder.__init__(
            self,
            name=name,
            pretrained_checkpoint_path=pretrained_checkpoint_path,
            data_norm_type=data_norm_type,
            patch_size=patch_size,
            vit_size=vit_size,
            pe_image_size=pe_image_size,
            torch_hub_force_reload=torch_hub_force_reload,
            gradient_checkpointing=gradient_checkpointing,
            keep_first_n_layers=keep_first_n_layers,
            use_pytorch_sdpa=use_pytorch_sdpa,
            disable_torch_compile_for_pe=disable_torch_compile_for_pe,
            *args,
            **kwargs,
        )

        IntermediateFeatureReturner.__init__(
            self,
            indices=indices,
            norm_intermediate=norm_intermediate,
        )

    def forward(self, encoder_input: ViTEncoderInput) -> List[ViTEncoderOutput]:
        """
        DUNE Encoder Forward Pass with Intermediate Feature Return

        Args:
            encoder_input (ViTEncoderInput): Input data for the encoder. Input data must contain image normalization type and normalized image tensor.

        Returns:
            List[ViTEncoderOutput]: Output data from the encoder. Returns a list of intermediate features.
        """
        # Check image normalization type
        self._check_data_normalization_type(encoder_input.data_norm_type)

        # Check the dtype and shape of the input image
        assert isinstance(encoder_input.image, torch.Tensor), "Input must be a torch.Tensor"
        assert encoder_input.image.ndim == 4, "Input must be of shape (B, C, H, W)"
        batch_size, channels, height, width = encoder_input.image.shape
        assert channels == 3, "Input must have 3 channels"
        assert (
            height % self.patch_size == 0 and width % self.patch_size == 0
        ), f"Input shape must be divisible by patch size: {self.patch_size}"

        if self.indices is None:
            self.indices = range(len(self.model.blocks))

        # Extract the intermediate features from the DINOv2 model
        intermediate_features = self.model.get_intermediate_layers(
            encoder_input.image,
            n=self.indices,
            reshape=True,
            norm=self.norm_intermediate,
        )

        # Convert the intermediate features to a list of ViTEncoderOutput
        intermediate_features = [ViTEncoderOutput(features=features) for features in intermediate_features]

        return intermediate_features


if __name__ == "__main__":
    # DUNE only has one variant: ViT-Base at 448 resolution
    # Checkpoint: https://download.europe.naverlabs.com/dune/dune_vitbase14_448.pth
    pretrained_checkpoint_path = "../../../checkpoints/encoders/dune_vitbase14_448.pth"

    # Init DUNE encoder
    dune_encoder = DUNEEncoder(
        name="dune_base",
        vit_size="base",
        pe_image_size=448,
        pretrained_checkpoint_path=pretrained_checkpoint_path,
    )

    print("DUNE Encoder has been initialized successfully!")

    # Intermediate Feature Returner Tests
    print("Running Intermediate Feature Returner Tests...")

    # Run the intermediate feature returner with last-n index
    dune_intermediate_feature_returner = DUNEIntermediateFeatureReturner(
        name="dune_base",
        vit_size="base",
        pe_image_size=448,
        pretrained_checkpoint_path=pretrained_checkpoint_path,
        indices=6,
    )  # Last 6 layers
    dummy_input = ViTEncoderInput(image=torch.randn(1, 3, 448, 448), data_norm_type="dune")
    output = dune_intermediate_feature_returner(dummy_input)
    assert isinstance(output, list), "Output must be a list of intermediate features"
    assert isinstance(output[0], ViTEncoderOutput), "Output must be a list of ViTEncoderOutput"
    assert len(output) == 6, "Output must have length of intermediate features equal to the number of indices"

    # Run the intermediate feature returner with specific indices
    dune_intermediate_feature_returner = DUNEIntermediateFeatureReturner(
        name="dune_base",
        vit_size="base",
        pe_image_size=448,
        pretrained_checkpoint_path=pretrained_checkpoint_path,
        indices=[0, 2, 4, 6],
    )  # Specific layers
    dummy_input = ViTEncoderInput(image=torch.randn(1, 3, 448, 448), data_norm_type="dune")
    output = dune_intermediate_feature_returner(dummy_input)
    assert isinstance(output, list), "Output must be a list of intermediate features"
    assert isinstance(output[0], ViTEncoderOutput), "Output must be a list of ViTEncoderOutput"
    assert len(output) == 4, "Output must have length of intermediate features equal to the number of indices"

    print("All Intermediate Feature Returner Tests have passed successfully!")



"""
Encoder Class for Meta Perception Encoder (PE)
"""

from typing import List, Optional, Union

import torch

import uniception.models.libs.perception_encoder.vision_encoder.pe as pe
from uniception.models.encoders import UniCeptionViTEncoderBase, ViTEncoderInput, ViTEncoderOutput
from uniception.models.utils.intermediate_feature_return import IntermediateFeatureReturner


class PerceptionEncoder(UniCeptionViTEncoderBase):
    "Transcribed implementation of the Perception Encoder"

    def __init__(
        self,
        name: str,
        data_norm_type: str = "perception_encoder",
        patch_size: int = 14,
        size: str = "large",
        checkpoint_type: str = "core",
        pretrained_checkpoint_path: str = None,
        gradient_checkpointing: bool = False,
        keep_first_n_layers: Optional[int] = None,
        *args,
        **kwargs,
    ):
        """
        Wrapper around the Perception Encoder.

        Args:
            name (str): Name of the encoder.
            data_norm_type (str): Type of data normalization to use.
            patch_size (int): Size of the patches to use.
            size (str): Size of the model, e.g., 'large'.
            checkpoint_type (str): Type of checkpoint to use, e.g., 'core'.
            pretrained_checkpoint_path (str, optional): Path to a pretrained checkpoint.
            gradient_checkpointing (bool): Whether to use gradient checkpointing.
            keep_first_n_layers (Optional[int]): Number of layers to keep from the original model.
        """
        # Init the base class
        super().__init__(
            name=name,
            data_norm_type=data_norm_type,
            size=size,
            patch_size=patch_size,
            gradient_checkpointing=gradient_checkpointing,
            *args,
            **kwargs,
        )

        # Init the encoder specific attributes
        self.enc_embed_dim = {"small": 384, "base": 768, "large": 1024, "giant": 1536}[size]

        # Define PE model factory
        model_ckpt_dict = {
            ("core", "base"): "PE-Core-B16-224",
            ("core", "large"): "PE-Core-L14-336",
            ("core", "giant"): "PE-Core-G14-448",
            ("lang", "base"): "PE-Lang-B16-224",
            ("lang", "large"): "PE-Lang-L14-448",
            ("lang", "giant"): "PE-Lang-G14-448",
            ("spatial", "giant"): "PE-Spatial-G14-448",
        }

        # Get the model type
        model_str = model_ckpt_dict.get((checkpoint_type, size), None)
        if model_str is None:
            raise ValueError(f"Unsupported combination of checkpoint_type '{checkpoint_type}' and size '{size}'.")

        # Init pre-trained model
        self.model = pe.VisionTransformer.from_config(model_str, pretrained=True)

        # Enable gradient checkpointing if required
        self.model.set_grad_checkpointing(gradient_checkpointing)

        # Keep only the first n layers of the model if keep_first_n_layers is specified
        if keep_first_n_layers is not None:
            self.model.truncate(keep_first_n_layers)

        # Load the custom pretrained checkpoint if provided
        if pretrained_checkpoint_path:
            print(f"Loading custom pretrained Perception Encoder checkpoint from {pretrained_checkpoint_path}")
            ckpt = torch.load(pretrained_checkpoint_path, weights_only=False)
            print(self.load_state_dict(ckpt["model"]))

    def forward(self, encoder_input: ViTEncoderInput) -> ViTEncoderOutput:
        """
        Perception Encoder Forward Pass

        Args:
            encoder_input (ViTEncoderInput): Input data for the encoder. Input data must contain image normalization type and normalized image tensor.

        Returns:
            ViTEncoderOutput: Output data from the encoder.
        """
        # Check image normalization type
        self._check_data_normalization_type(encoder_input.data_norm_type)

        # Check the dtype and shape of the input image
        assert isinstance(encoder_input.image, torch.Tensor), "Input must be a torch.Tensor"
        assert encoder_input.image.ndim == 4, "Input must be of shape (B, C, H, W)"
        batch_size, channels, height, width = encoder_input.image.shape
        assert channels == 3, "Input must have 3 channels"
        assert (
            height % self.patch_size == 0 and width % self.patch_size == 0
        ), f"Input shape must be divisible by patch size: {self.patch_size}"

        # Extract the features from the PE model
        features = self.model.forward_features(
            encoder_input.image,
            norm=True,
            strip_cls_token=True,  # Strip the CLS token if it exists
        )

        # Permute the features to match the expected output shape
        features = features.permute(0, 2, 1).reshape(
            batch_size, -1, height // self.patch_size, width // self.patch_size
        )

        return ViTEncoderOutput(features=features)


class PerceptionEncoderIntermediateFeatureReturner(PerceptionEncoder, IntermediateFeatureReturner):
    "Intermediate Feature Returner for UniCeption Perception Encoder"

    def __init__(
        self,
        name: str,
        data_norm_type: str = "perception_encoder",
        patch_size: int = 14,
        size: str = "large",
        checkpoint_type: str = "core",
        pretrained_checkpoint_path: str = None,
        gradient_checkpointing: bool = False,
        keep_first_n_layers: Optional[int] = None,
        indices: Optional[Union[int, List[int]]] = 1,
        norm_intermediate: bool = True,
        *args,
        **kwargs,
    ):
        """
        Preception Encoder for extracting intermediate features.

        Args:
            name (str): Name of the encoder.
            data_norm_type (str): Type of data normalization to use.
            patch_size (int): Size of the patches to use.
            size (str): Size of the model, e.g., 'large'.
            checkpoint_type (str): Type of checkpoint to use, e.g., 'core'.
            pretrained_checkpoint_path (str, optional): Path to a pretrained checkpoint.
            gradient_checkpointing (bool): Whether to use gradient checkpointing.
            keep_first_n_layers (Optional[int]): Number of layers to keep from the original model.
            indices (Optional[Union[int, List[int]]], optional): Indices of the layers to return. Defaults to 1. Options:
            - int: Return the last n layers.
            - List[int]: Return the intermediate layers at the specified indices.
            norm_intermediate (bool, optional): Whether to normalize the intermediate features. Defaults to True.
        """
        # Init the base classes
        PerceptionEncoder.__init__(
            self,
            name=name,
            data_norm_type=data_norm_type,
            patch_size=patch_size,
            size=size,
            checkpoint_type=checkpoint_type,
            pretrained_checkpoint_path=pretrained_checkpoint_path,
            gradient_checkpointing=gradient_checkpointing,
            keep_first_n_layers=keep_first_n_layers,
            *args,
            **kwargs,
        )
        IntermediateFeatureReturner.__init__(
            self,
            indices=indices,
            norm_intermediate=norm_intermediate,
        )

    def forward(self, encoder_input: ViTEncoderInput) -> List[ViTEncoderOutput]:
        """
        Forward pass for the Perception Encoder with intermediate feature return.

        Args:
            encoder_input (ViTEncoderInput): Input data for the encoder. Input data must contain image normalization type and normalized image tensor.

        Returns:
            List[ViTEncoderOutput]: List of outputs from the encoder at specified indices.
        """
        # Check image normalization type
        self._check_data_normalization_type(encoder_input.data_norm_type)

        # Check the dtype and shape of the input image
        assert isinstance(encoder_input.image, torch.Tensor), "Input must be a torch.Tensor"
        assert encoder_input.image.ndim == 4, "Input must be of shape (B, C, H, W)"
        batch_size, channels, height, width = encoder_input.image.shape
        assert channels == 3, "Input must have 3 channels"
        assert (
            height % self.patch_size == 0 and width % self.patch_size == 0
        ), f"Input shape must be divisible by patch size: {self.patch_size}"

        if self.indices is None:
            self.indices = range(len(self.model.transformer.resblocks))

        # Extract the intermediate features from the PE model
        intermediate_feats = self.model.get_intermediate_layers(
            x=encoder_input.image,
            n=self.indices,
            norm=self.norm_intermediate,
            strip_cls_token=True,  # Strip the CLS token if it exists
        )

        # reshape the features to match the expected output shape
        outputs = []

        Hp, Wp = height // self.patch_size, width // self.patch_size
        outputs = [
            ViTEncoderOutput(features=feat.permute(0, 2, 1).reshape(batch_size, -1, Hp, Wp))
            for feat in intermediate_feats
        ]

        return outputs


if __name__ == "__main__":
    # Example usage
    encoder = PerceptionEncoderIntermediateFeatureReturner(
        name="PerceptionEncoder", size="giant", checkpoint_type="spatial", indices=[0, 1, 2, 3]
    ).cuda()

    dummy_input = ViTEncoderInput(image=torch.randn(1, 3, 448, 448).cuda(), data_norm_type="perception_encoder")

    output = encoder(dummy_input)

    for x in output:
        print(x.features.shape)  # Should print the shape of the output features



"""
Script to convert DINOv2 weights to UniCeption info sharing classes.
"""

import argparse
import os

import torch

from uniception.models.encoders import _make_encoder_test
from uniception.models.encoders.base import ViTEncoderInput
from uniception.models.info_sharing.alternating_attention_transformer import MultiViewAlternatingAttentionTransformer
from uniception.models.info_sharing.base import MultiViewTransformerInput
from uniception.models.info_sharing.global_attention_transformer import MultiViewGlobalAttentionTransformer
from uniception.models.utils.transformer_blocks import Mlp, SwiGLUFFNFused


def get_parser():
    parser = argparse.ArgumentParser(description="Convert DINOv2 weights for info sharing classes.")
    parser.add_argument(
        "--encoder_str", type=str, default="dinov2_large", help="DINOv2 encoder string (e.g., dinov2_large)."
    )
    parser.add_argument("--start", type=int, default=12, help="Starting block index (including)")
    parser.add_argument(
        "--end", type=int, default=-1, help="Ending block index (excluding), use -1 to indicate the last block"
    )
    parser.add_argument(
        "--info_sharing_layers",
        type=int,
        default=-1,
        help="Depth of info sharing transformer. (-1) to determine from start and end. \
        If different then (to - from), will left empty layers or not copy all layers from the encoder.",
    )
    parser.add_argument("--info_sharing_class", type=str, default="global", help="'global' or 'alternating'")
    parser.add_argument("--output_path", type=str, help="Path to save the converted info sharing model.")
    parser.add_argument(
        "--skip_verify", action="store_true", help="Whether to skip verifying the converted model by loading it back."
    )
    return parser


def main(args):
    # Construct UniCeption encoder
    print(f"Loading encoder: {args.encoder_str}")
    encoder = _make_encoder_test(args.encoder_str)

    # Decide the length of layers to extract
    encoder_length = len(encoder.model.blocks)
    if args.end == -1:
        args.end = encoder_length

    layers_to_extract = list(range(args.start, args.end))
    layers_to_extract = layers_to_extract[: min(len(layers_to_extract), encoder_length)]

    if args.info_sharing_layers == -1:
        args.info_sharing_layers = len(layers_to_extract)

    layers_to_extract = layers_to_extract[: min(len(layers_to_extract), args.info_sharing_layers)]

    print("-" * 40)
    print(
        f"Extracting blocks from {layers_to_extract[0]} to {layers_to_extract[-1] + 1} (encoder length: {encoder_length})"
    )
    print(f"Filling into {len(layers_to_extract)}/{args.info_sharing_layers} info sharing layers")

    # Determine the MLP class based on the encoder
    if "giant" in args.encoder_str:
        mlp_layer = SwiGLUFFNFused
    else:
        mlp_layer = Mlp

    # Construct info sharing model
    target_blocks: torch.nn.ModuleList

    if args.info_sharing_class == "global":
        info_sharing_model = MultiViewGlobalAttentionTransformer(
            name="info_sharing",
            input_embed_dim=encoder.model.embed_dim,
            distinguish_ref_and_non_ref_views=False,  # for ease of validation
            use_pe_for_non_reference_views=False,
            max_num_views_for_pe=1000,  # dummy value
            use_rand_idx_pe_for_non_reference_views=False,  # dummy value
            depth=args.info_sharing_layers,
            dim=encoder.model.embed_dim,
            num_heads=encoder.model.num_heads,
            mlp_layer=mlp_layer,
            mlp_ratio=4,  # the same for all DINOv2
            qkv_bias=True,
            qk_norm=False,
            init_values=1e-5,  # for enabling LayerScale
        )

        target_blocks = info_sharing_model.self_attention_blocks
    elif args.info_sharing_class == "alternating":
        info_sharing_model = MultiViewAlternatingAttentionTransformer(
            name="info_sharing",
            input_embed_dim=encoder.model.embed_dim,
            distinguish_ref_and_non_ref_views=False,  # for ease of validation
            use_pe_for_non_reference_views=False,
            max_num_views_for_pe=1000,  # dummy value
            use_rand_idx_pe_for_non_reference_views=False,  # dummy value
            depth=args.info_sharing_layers,
            dim=encoder.model.embed_dim,
            num_heads=encoder.model.num_heads,
            mlp_layer=mlp_layer,
            mlp_ratio=4,  # the same for all DINOv2
            qkv_bias=True,
            qk_norm=False,
            init_values=1e-5,  # for enabling LayerScale
        )

        target_blocks = info_sharing_model.self_attention_blocks
    else:
        raise ValueError(f"Unknown info sharing class: {args.info_sharing_class}")

    # Fill in the weights
    for i, layer_idx in enumerate(layers_to_extract):
        print(f"Copying encoder block {layer_idx} to info sharing block {i}")
        target_blocks[i].load_state_dict(encoder.model.blocks[layer_idx].state_dict())

    # Also copy weight for the final layernorm
    print("Copying encoder final norm to info sharing norm")
    info_sharing_model.norm.load_state_dict(encoder.model.norm.state_dict())

    # Save the converted model
    print(f"Saving to {args.output_path}")
    os.makedirs(os.path.dirname(args.output_path), exist_ok=True)
    torch.save({"model": info_sharing_model.state_dict()}, args.output_path)


def verify(args):
    print("Verifying the converted model...")

    # Construct original encoder
    encoder_reference = _make_encoder_test(
        args.encoder_str,
        norm_returned_features=True,  # keep final norm for reference encoder
    )
    encoder_test = _make_encoder_test(
        args.encoder_str,
        norm_returned_features=False,  # drop final norm for truncated encoder
        keep_first_n_layers=args.start,  # keep only the first 'start' layers
    )

    encoder_length = len(encoder_reference.model.blocks)
    if args.end == -1:
        args.end = encoder_length

    layers_to_extract = list(range(args.start, args.end))
    layers_to_extract = layers_to_extract[: min(len(layers_to_extract), encoder_length)]

    if args.info_sharing_layers == -1:
        args.info_sharing_layers = len(layers_to_extract)

    assert args.start > 0, "Verification only works for start > 0"
    assert args.info_sharing_layers == (
        args.end - args.start
    ), "Verification only works for info_sharing_layers == (end - start)"
    assert args.end <= len(encoder_reference.model.blocks), "Verification only works for end <= encoder length"

    encoder_reference.model.blocks = encoder_reference.model.blocks[: args.end]  # keep the full encoder up to "end"

    # Determine the MLP class based on the encoder
    if "giant" in args.encoder_str:
        mlp_layer = SwiGLUFFNFused
    else:
        mlp_layer = Mlp

    # Construct info sharing model
    info_sharing_model: torch.nn.Module
    if args.info_sharing_class == "global":
        info_sharing_model = MultiViewGlobalAttentionTransformer(
            name="info_sharing",
            input_embed_dim=encoder_reference.model.embed_dim,
            distinguish_ref_and_non_ref_views=False,  # for ease of validation
            use_pe_for_non_reference_views=False,  # necessary for preserving DINO features
            max_num_views_for_pe=1000,  # dummy value
            use_rand_idx_pe_for_non_reference_views=False,  # dummy value
            depth=args.info_sharing_layers,
            dim=encoder_reference.model.embed_dim,
            num_heads=encoder_reference.model.num_heads,
            mlp_layer=mlp_layer,
            mlp_ratio=4,  # the same for all DINOv2
            qkv_bias=True,
            qk_norm=False,
            init_values=1e-5,  # for enabling LayerScale
            pretrained_checkpoint_path=args.output_path,
        )
    elif args.info_sharing_class == "alternating":
        info_sharing_model = MultiViewAlternatingAttentionTransformer(
            name="info_sharing",
            input_embed_dim=encoder_reference.model.embed_dim,
            distinguish_ref_and_non_ref_views=False,  # for ease of validation
            use_pe_for_non_reference_views=False,  # necessary for preserving DINO features
            max_num_views_for_pe=1000,  # dummy value
            use_rand_idx_pe_for_non_reference_views=False,  # dummy value
            depth=args.info_sharing_layers,
            dim=encoder_reference.model.embed_dim,
            num_heads=encoder_reference.model.num_heads,
            mlp_layer=mlp_layer,
            mlp_ratio=4,  # the same for all DINOv2
            qkv_bias=True,
            qk_norm=False,
            init_values=1e-5,  # for enabling LayerScale
            pretrained_checkpoint_path=args.output_path,
        )
    else:
        raise ValueError(f"Unknown info sharing class: {args.info_sharing_class}")

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # Move models to device
    encoder_reference = encoder_reference.to(device)
    encoder_test = encoder_test.to(device)
    info_sharing_model = info_sharing_model.to(device)

    torch.manual_seed(42)
    dummy_input = torch.randn(1, 3, 224, 224).to(device)

    with torch.no_grad():
        # Forward pass through the reference encoder
        reference_feature = encoder_reference(ViTEncoderInput(image=dummy_input, data_norm_type="dinov2")).features

        # Forward pass through the test encoder + info sharing(without any PE in info_sharing)
        # Pass through the truncated encoder
        encoder_result = encoder_test(ViTEncoderInput(image=dummy_input, data_norm_type="dinov2"))

        features_test_encoder = encoder_result.features
        registers = encoder_result.registers

        # Passing in the registers is ABSOLUTELY ESSENTIAL for correct dino feature to appear in the info-sharing module.
        # if not passed, would result in 1.15 relative error! which totally changes the feature distribution.
        info_sharing_output = info_sharing_model(
            MultiViewTransformerInput(
                features=[features_test_encoder], additional_input_tokens_per_view=[registers]  # single view
            )
        ).features[0]

        relative_error = torch.linalg.norm(info_sharing_output - reference_feature) / torch.linalg.norm(
            reference_feature
        )
        assert relative_error < 1e-3, f"Verification failed: Relative error {relative_error} is too high!"
        print(
            f"Relative error between {len(encoder_reference.model.blocks)} layer encoder to {len(encoder_test.model.blocks)} layer encoder + {len(info_sharing_model.self_attention_blocks)} layer info-sharing: {relative_error:.3e}"
        )

    print("Verification successful: The features from the reference encoder and the encoder + info sharing match!")


if __name__ == "__main__":
    parser = get_parser()
    args = parser.parse_args()
    main(args)

    if not args.skip_verify:
        verify(args)



"""
This file compares the speed of transformer engine v.s. UniCeption Attention Implementation.
"""

import torch
import transformer_engine.pytorch as te
from transformer_engine.common.recipe import Format, DelayedScaling

from uniception.models.utils.transformer_blocks import SelfAttentionBlock


def cuda_profile(profile_func, warmup: int = 10, test: int = 100):
    """
    Run a CUDA‑backed callable several times and report timing statistics.

    Parameters
    ----------
    profile_func : Callable[[], Any]
        A function that launches the CUDA work you want to measure.
        It should *not* include its own synchronizations.
    warmup : int
        Number of warm‑up iterations to ignore in the timing results.
    test : int
        Number of timed iterations.

    Returns
    -------
    dict
        {
            'avg_ms':   mean latency per iteration (milliseconds),
            'std_ms':   un‑biased std‑dev of latency (milliseconds),
            'total_ms': total measured wall‑clock time (milliseconds),
            'iters_per_sec': throughput in iterations / second
        }
    """
    if not torch.cuda.is_available():
        raise RuntimeError("CUDA is not available on this system.")

    # Ensure the first kernel launch won’t be timed
    torch.cuda.synchronize()

    # -------------------
    # Warm‑up phase
    # -------------------
    for _ in range(warmup):
        profile_func()
    torch.cuda.synchronize()

    # -------------------
    # Timed phase
    # -------------------
    start_evt = torch.cuda.Event(enable_timing=True)
    end_evt = torch.cuda.Event(enable_timing=True)
    times_ms = torch.empty(test, dtype=torch.float64, device="cpu")

    for i in range(test):
        start_evt.record()
        profile_func()
        end_evt.record()
        torch.cuda.synchronize()  # wait for kernels
        times_ms[i] = start_evt.elapsed_time(end_evt)  # ms for this iter

    # -------------------
    # Aggregate stats
    # -------------------
    avg = times_ms.mean().item()
    std = times_ms.std(unbiased=True).item()
    total = times_ms.sum().item()

    return {
        "avg_ms": avg,
        "std_ms": std,
        "total_ms": total,
        "iters_per_sec": 1_000.0 / avg if avg > 0 else float("inf"),
    }


def te_fused(batch_size: int, head_dim: int, n_head: int, seq_len: int, precision: str = "bfloat16"):
    """
    Profile the Transformer Engine fused kernel
    """

    te_transformer = (
        te.TransformerLayer(
            head_dim * n_head,
            4 * head_dim * n_head,
            n_head,
            fuse_qkv_params=True,
            self_attn_mask_type="no_mask",
            hidden_dropout=0,
            attention_dropout=0,
            window_size=(-1, -1),
        )
        .cuda()
        .to(dtype=torch.bfloat16)
    )

    seq_len_list = [
        seq_len,
        int(1.25 * seq_len),
        int(1.5 * seq_len),
        int(1.75 * seq_len),
        int(2.0 * seq_len),
        int(0.75 * seq_len),
        int(0.5 * seq_len),
    ]
    x_list = [torch.rand(s, batch_size, head_dim * n_head).cuda().to(dtype=torch.bfloat16) for s in seq_len_list]

    if precision == "bfloat16":

        def profile_func():
            for x in x_list:
                y = te_transformer(x)
                torch.sum(y).backward()

    elif precision == "fp8":
        fp8_format = Format.HYBRID
        fp8_recipe = DelayedScaling(fp8_format=fp8_format, amax_history_len=16, amax_compute_algo="max")

        def profile_func():
            with te.fp8_autocast(enabled=True, fp8_recipe=fp8_recipe):
                for x in x_list:
                    y = te_transformer(x)
                    torch.sum(y).backward()

    result = cuda_profile(profile_func, 5, 100)

    return result


def uniception(batch_size: int, head_dim: int, n_head: int, seq_len: int, precision: str = "bfloat16"):

    attention = torch.compile(
        SelfAttentionBlock(dim=head_dim * n_head, num_heads=n_head, qkv_bias=True).cuda().to(dtype=torch.bfloat16),
        dynamic=True,
        fullgraph=True,
    )

    seq_len_list = [
        seq_len,
        int(1.25 * seq_len),
        int(1.5 * seq_len),
        int(1.75 * seq_len),
        int(2.0 * seq_len),
        int(0.75 * seq_len),
        int(0.5 * seq_len),
    ]
    x_list = [torch.rand(s, batch_size, head_dim * n_head).cuda().to(dtype=torch.bfloat16) for s in seq_len_list]

    def profile_func():
        for x in x_list:
            y = attention(x)
            torch.sum(y).backward()

    result = cuda_profile(profile_func, 5, 100)

    return result


if __name__ == "__main__":

    import pandas as pd

    dataframe_data = []

    # batch_sizes = [16, 32, 64, 128, 256]
    # seq_lengths = [512, 1024, 2048, 3072]

    batch_sizes = [16]
    seq_lengths = [512]

    for bs in batch_sizes:
        for seq_len in seq_lengths:
            # test te bfloat16
            results_te_fused = te_fused(bs, 128, 8, seq_len)
            results_te_fused["method"] = "te"
            results_te_fused["precision"] = "bfloat16"
            results_te_fused["batch_size"] = bs
            results_te_fused["seq_length"] = seq_len

            dataframe_data.append(results_te_fused)

            # test UniCeption bfloat16
            results_uc = uniception(bs, 128, 8, seq_len)
            results_uc["method"] = "uniception"
            results_uc["precision"] = "bfloat16"
            results_uc["batch_size"] = bs
            results_uc["seq_length"] = seq_len

            dataframe_data.append(results_uc)

            # test te fp8
            results_te_fused_fp8 = te_fused(bs, 128, 8, seq_len, precision="fp8")
            results_te_fused_fp8["method"] = "te"
            results_te_fused_fp8["precision"] = "fp8"
            results_te_fused_fp8["batch_size"] = bs
            results_te_fused_fp8["seq_length"] = seq_len

            dataframe_data.append(results_te_fused_fp8)

    dataframe = pd.DataFrame(dataframe_data)
    dataframe.to_csv("dataframe.csv", index=False)



import os

os.environ["TORCH_LOGS"] = "recompiles"
# os.environ["PYTORCH_JIT_LOG_SCHEMA"] = "1"

import torch
from uniception.models.encoders import encoder_factory, ViTEncoderInput
from itertools import product

if __name__ == "__main__":

    # Test the dinov2 encoder
    model = encoder_factory("dinov2", size="large", name="encoder", disable_torch_compile_for_pe=True).cuda()
    model.compile(
        dynamic=True,
        fullgraph=False,
        options={
            "shape_padding": True,
        },
    )

    print(model)

    # Create a dummy input tensor
    hw_list = [(int(x * 14), int(y * 14)) for (x, y) in product(range(16, 24), range(16, 24))]

    dummy_input = [ViTEncoderInput(image=torch.randn(16, 3, *hw).cuda(), data_norm_type="dinov2") for hw in hw_list]

    for i in range(100):
        # Forward pass through the model
        for x in dummy_input:
            with torch.autocast("cuda", dtype=torch.bfloat16):
                output = model(x)
                print(x.image.shape)



"""
Download the UniCeption format checkpoints from the AirLab Data Server
"""

import argparse
import os

import urllib3
from minio import Minio
from minio.error import S3Error
from tqdm import tqdm


def main():
    parser = argparse.ArgumentParser(description="Download UniCeption format checkpoints from AirLab Data Server")
    parser.add_argument(
        "--folders",
        nargs="+",
        default=["encoders", "info_sharing", "prediction_heads", "examples"],
        help="List of folders to download (default: all folders). Choices: encoders, info_sharing, prediction_heads, examples",
    )
    parser.add_argument("--destination", type=str, default="./", help="Destination folder for downloaded checkpoints")
    args = parser.parse_args()

    access_key = "bT79gQYtfhpxFIitlpns"
    secret_key = "g7mSvUJ5k2a9mKv9IbhwXmUQjQX52MLwulhW9ONO"
    # client = Minio("airlab-share-02.andrew.cmu.edu:9000", access_key=access_key, secret_key=secret_key, secure=True)
    client = Minio(
        "airlab-share-02.andrew.cmu.edu:9000",
        access_key=access_key,
        secret_key=secret_key,
        secure=True,
        http_client=urllib3.PoolManager(cert_reqs="CERT_NONE"),  # disables SSL verification
    )

    bucket_name = "uniception"

    def download_folder(folder_name, bucket_name, client, destination_folder):
        folder_name = f"checkpoints/{folder_name}/"
        objects = client.list_objects(bucket_name, prefix=folder_name, recursive=True)
        for obj in tqdm(objects, desc=f"Downloading {folder_name}"):
            destination_file = os.path.join(destination_folder, obj.object_name)
            if not os.path.exists(destination_file):
                os.makedirs(os.path.dirname(destination_file), exist_ok=True)
                try:
                    client.fget_object(bucket_name, obj.object_name, destination_file)
                    print(f"Downloaded {obj.object_name} to {destination_file}")
                except S3Error as e:
                    print(f"Error downloading {obj.object_name}: {e}")
            else:
                print(f"File {destination_file} already exists. Skipping...")

    for folder in args.folders:
        download_folder(folder, bucket_name, client, args.destination)


if __name__ == "__main__":
    main()



"""
Encoder Class for DINOv3 - ViT & ConvNext Variants
"""

import os
from typing import List, Optional, Union

import torch
import torch.nn as nn

from uniception.models.encoders.base import (
    UniCeptionEncoderBase,
    UniCeptionViTEncoderBase,
    ViTEncoderInput,
    ViTEncoderOutput,
)
from uniception.models.utils.intermediate_feature_return import IntermediateFeatureReturner


class DINOv3Encoder(UniCeptionViTEncoderBase):
    "UniCeption DINOv3 ViT Encoder"

    def __init__(
        self,
        name: str,
        dinov3_repo_dir: str,
        data_norm_type: str = "dinov3",
        patch_size: int = 16,
        size: str = "large",
        weights: Optional[str] = None,
        pretrained_checkpoint_path: str = None,
        gradient_checkpointing: bool = False,
        keep_first_n_layers: Optional[int] = None,
        disable_torch_compile_for_pe=False,
        *args,
        **kwargs,
    ):
        """
        DINOv3 ViT Encoders for extracting spatial features from images.

        Args:
            name (str): Name of the encoder.
            dinov3_repo_dir (str): Path to the local directory where the DINOv3 repo was cloned. Required.
            data_norm_type (str): Image normalization type. Default: "dinov3"
            patch_size (int): Patch size for the encoder. Default: 16
            size (str): Size variant of the DINOv3 model. Options: ["small", "small+", "base", "large", "huge+", "7b"]. Default: "large"
            weights (Optional[str]): Path or URL to pretrained DINOv3 backbone checkpoint to pass to torch.hub.load. Default: None
            pretrained_checkpoint_path (str): Path to the pretrained checkpoint if using custom trained version of DINOv3. Default: None
            gradient_checkpointing (bool): Whether to use gradient checkpointing to save GPU memory during backward call. Default: False
            keep_first_n_layers (Optional[int]): If specified, only the first n layers of the model will be kept. Default: None
            disable_torch_compile_for_pe (bool): Whether to disable torch compile for PE interpolation. Default: False
        """
        # Init the base class
        super().__init__(
            name=name,
            data_norm_type=data_norm_type,
            patch_size=patch_size,
            gradient_checkpointing=gradient_checkpointing,
            *args,
            **kwargs,
        )

        # Init the DINOv3 Encoder specific attributes
        self.version = size
        self.enc_embed_dim = {
            "small": 384,
            "small+": 384,
            "base": 768,
            "large": 1024,
            "huge+": 1280,
            "7b": 4096,
        }[self.version]

        # Define DINOv3 model factory
        DINO_MODELS = {
            "small": "dinov3_vits16",
            "small+": "dinov3_vits16plus",
            "base": "dinov3_vitb16",
            "large": "dinov3_vitl16",
            "huge+": "dinov3_vith16plus",
            "7b": "dinov3_vit7b16",
        }

        # Load the pretrained DINOv3 model from local repo
        print(f"Loading {DINO_MODELS[self.version]} from local repo: {dinov3_repo_dir}")
        if weights:
            print(f"Using pretrained weights from: {weights}")
            self.model = torch.hub.load(
                dinov3_repo_dir,
                DINO_MODELS[self.version],
                source="local",
                weights=weights,
            )
        else:
            self.model = torch.hub.load(
                dinov3_repo_dir,
                DINO_MODELS[self.version],
                source="local",
                pretrained=False,
            )

        # Disable position interpolation at PE interpolation to enable torch compile
        # when training with multiple image shapes.
        if disable_torch_compile_for_pe:
            self.model.interpolate_pos_encoding = torch.compiler.disable(self.model.interpolate_pos_encoding)

        # Keep only the first n layers of the model if keep_first_n_layers is specified
        if keep_first_n_layers is not None:
            self.model.blocks = nn.ModuleList(self.model.blocks[:keep_first_n_layers])

        # Wrap the transformer blocks with support for gradient checkpointing if required
        if self.gradient_checkpointing:
            for i in range(len(self.model.blocks)):
                self.model.blocks[i] = self.wrap_module_with_gradient_checkpointing(self.model.blocks[i])

        # Load the custom pretrained checkpoint if provided
        if pretrained_checkpoint_path:
            print(f"Loading custom pretrained DINOv3 checkpoint from {pretrained_checkpoint_path}")
            ckpt = torch.load(pretrained_checkpoint_path, weights_only=False)
            print(self.load_state_dict(ckpt["model"]))

    def forward(self, encoder_input: ViTEncoderInput) -> ViTEncoderOutput:
        """
        DINOv3 Encoder Forward Pass

        Args:
            encoder_input (ViTEncoderInput): Input data for the encoder. Input data must contain image normalization type and normalized image tensor.

        Returns:
            ViTEncoderOutput: Output data from the encoder.
        """
        # Check image normalization type
        self._check_data_normalization_type(encoder_input.data_norm_type)

        # Check the dtype and shape of the input image
        assert isinstance(encoder_input.image, torch.Tensor), "Input must be a torch.Tensor"
        assert encoder_input.image.ndim == 4, "Input must be of shape (B, C, H, W)"
        batch_size, channels, height, width = encoder_input.image.shape
        assert channels == 3, "Input must have 3 channels"
        assert (
            height % self.patch_size == 0 and width % self.patch_size == 0
        ), f"Input shape must be divisible by patch size: {self.patch_size}"

        # Extract the features from the DINOv3 model
        features = self.model.forward_features(encoder_input.image)["x_norm_patchtokens"]

        # Resize the features to the expected shape
        # (B x Num_patches x Embed_dim) -> (B x Embed_dim x H / Patch_Size x W / Patch_Size)
        features = features.permute(0, 2, 1)
        features = features.reshape(
            -1, self.enc_embed_dim, height // self.patch_size, width // self.patch_size
        ).contiguous()

        return ViTEncoderOutput(features=features)


class DINOv3IntermediateFeatureReturner(DINOv3Encoder, IntermediateFeatureReturner):
    "Intermediate Feature Returner for UniCeption DINOv3 ViT Encoder"

    def __init__(
        self,
        name: str,
        dinov3_repo_dir: str,
        data_norm_type: str = "dinov3",
        patch_size: int = 16,
        size: str = "large",
        weights: Optional[str] = None,
        pretrained_checkpoint_path: str = None,
        gradient_checkpointing: bool = False,
        keep_first_n_layers: Optional[int] = None,
        disable_torch_compile_for_pe=False,
        indices: Optional[Union[int, List[int]]] = 1,
        norm_intermediate: bool = True,
        *args,
        **kwargs,
    ):
        """
        DINOv3 Encoder for extracting spatial features from images.

        Args:
            name (str): Name of the encoder.
            dinov3_repo_dir (str): Path to the local directory where the DINOv3 repo was cloned. Required.
            data_norm_type (str): Image normalization type. Default: "dinov3"
            patch_size (int): Patch size for the encoder. Default: 16
            size (str): Size variant of the DINOv3 model. Options: ["small", "small+", "base", "large", "huge+", "7b"]. Default: "large"
            weights (Optional[str]): Path or URL to pretrained DINOv3 backbone checkpoint to pass to torch.hub.load. Default: None
            pretrained_checkpoint_path (str): Path to the pretrained checkpoint if using custom trained version of DINOv3. Default: None
            gradient_checkpointing (bool): Whether to use gradient checkpointing to save GPU memory during backward call. Default: False
            keep_first_n_layers (Optional[int]): If specified, only the first n layers of the model will be kept. Default: None
            disable_torch_compile_for_pe (bool): Whether to disable torch compile for PE interpolation. Default: False
            indices (Optional[Union[int, List[int]]], optional): Indices of the layers to return. Defaults to 1. Options:
            - int: Return the last n layers.
            - List[int]: Return the intermediate layers at the specified indices.
            norm_intermediate (bool, optional): Whether to normalize the intermediate features. Defaults to True.
        """
        # Init the base classes
        DINOv3Encoder.__init__(
            self,
            name=name,
            data_norm_type=data_norm_type,
            patch_size=patch_size,
            size=size,
            dinov3_repo_dir=dinov3_repo_dir,
            weights=weights,
            pretrained_checkpoint_path=pretrained_checkpoint_path,
            gradient_checkpointing=gradient_checkpointing,
            keep_first_n_layers=keep_first_n_layers,
            disable_torch_compile_for_pe=disable_torch_compile_for_pe,
            *args,
            **kwargs,
        )
        IntermediateFeatureReturner.__init__(
            self,
            indices=indices,
            norm_intermediate=norm_intermediate,
        )

    def forward(self, encoder_input: ViTEncoderInput) -> List[ViTEncoderOutput]:
        """
        DINOv3 Encoder Forward Pass with Intermediate Feature Return

        Args:
            encoder_input (ViTEncoderInput): Input data for the encoder. Input data must contain image normalization type and normalized image tensor.

        Returns:
            List[ViTEncoderOutput]: Output data from the encoder. Returns a list of intermediate features.
        """
        # Check image normalization type
        self._check_data_normalization_type(encoder_input.data_norm_type)

        # Check the dtype and shape of the input image
        assert isinstance(encoder_input.image, torch.Tensor), "Input must be a torch.Tensor"
        assert encoder_input.image.ndim == 4, "Input must be of shape (B, C, H, W)"
        batch_size, channels, height, width = encoder_input.image.shape
        assert channels == 3, "Input must have 3 channels"
        assert (
            height % self.patch_size == 0 and width % self.patch_size == 0
        ), f"Input shape must be divisible by patch size: {self.patch_size}"

        if self.indices is None:
            self.indices = range(len(self.model.blocks))

        # Extract the intermediate features from the DINOv3 model
        intermediate_features = self.model.get_intermediate_layers(
            encoder_input.image,
            n=self.indices,
            reshape=True,
            norm=self.norm_intermediate,
        )

        # Convert the intermediate features to a list of ViTEncoderOutput
        intermediate_features = [ViTEncoderOutput(features=features) for features in intermediate_features]

        return intermediate_features


class DINOv3ConvNextEncoder(UniCeptionEncoderBase):
    "UniCeption DINOv3 ConvNext Encoder"

    def __init__(
        self,
        name: str,
        dinov3_repo_dir: str,
        data_norm_type: str = "dinov3",
        patch_size: int = 16,
        size: str = "large",
        weights: Optional[str] = None,
        pretrained_checkpoint_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        """
        DINOv3 ConvNext Encoders for extracting spatial features from images.

        Args:
            name (str): Name of the encoder.
            dinov3_repo_dir (str): Path to the local directory where the DINOv3 repo was cloned. Required.
            data_norm_type (str): Image normalization type. Default: "dinov3"
            patch_size (int): Patch size for the encoder. Default: 16
            size (str): Size variant of the DINOv3 model. Options: ["tiny", "small", "base", "large"]. Default: "large"
            weights (Optional[str]): Path or URL to pretrained DINOv3 backbone checkpoint to pass to torch.hub.load. Default: None
            pretrained_checkpoint_path (str): Path to the pretrained checkpoint if using custom trained version of DINOv3. Default: None
        """
        # Init the base class
        super().__init__(name=name, data_norm_type=data_norm_type, *args, **kwargs)

        # Init the DINOv3 ConvNext Encoder specific attributes
        self.patch_size = patch_size
        self.final_layer_patch_size = 32  # DINOv3 ConvNext uses a final patch size of 32
        self.version = size
        self.enc_embed_dim = {
            "tiny": 768,
            "small": 768,
            "base": 1024,
            "large": 1536,
        }[self.version]

        # Define DINOv3 ConvNext factory
        DINO_CONV_MODELS = {
            "tiny": "dinov3_convnext_tiny",
            "small": "dinov3_convnext_small",
            "base": "dinov3_convnext_base",
            "large": "dinov3_convnext_large",
        }

        # Load the pretrained DINOv3 model from local repo
        print(f"Loading {DINO_CONV_MODELS[self.version]} from local repo: {dinov3_repo_dir}")
        if weights:
            print(f"Using pretrained weights from: {weights}")
            self.model = torch.hub.load(
                dinov3_repo_dir,
                DINO_CONV_MODELS[self.version],
                source="local",
                weights=weights,
            )
        else:
            self.model = torch.hub.load(
                dinov3_repo_dir,
                DINO_CONV_MODELS[self.version],
                source="local",
                pretrained=False,
            )

        # Init the patch size arg for the model
        self.model.patch_size = self.patch_size

        # Load the custom pretrained checkpoint if provided
        if pretrained_checkpoint_path:
            print(f"Loading custom pretrained DINOv3 checkpoint from {pretrained_checkpoint_path}")
            ckpt = torch.load(pretrained_checkpoint_path, weights_only=False)
            print(self.load_state_dict(ckpt["model"]))

    def forward(self, encoder_input: ViTEncoderInput) -> ViTEncoderOutput:
        """
        DINOv3 Encoder Forward Pass

        Args:
            encoder_input (ViTEncoderInput): Input data for the encoder. Input data must contain image normalization type and normalized image tensor.

        Returns:
            ViTEncoderOutput: Output data from the encoder.
        """
        # Check image normalization type
        self._check_data_normalization_type(encoder_input.data_norm_type)

        # Check the dtype and shape of the input image
        assert isinstance(encoder_input.image, torch.Tensor), "Input must be a torch.Tensor"
        assert encoder_input.image.ndim == 4, "Input must be of shape (B, C, H, W)"
        batch_size, channels, height, width = encoder_input.image.shape
        assert channels == 3, "Input must have 3 channels"
        assert (
            height % self.patch_size == 0 and width % self.patch_size == 0
        ), f"Input shape must be divisible by patch size: {self.patch_size}"

        # Extract the features from the DINOv3 model
        features = self.model.forward_features(encoder_input.image)["x_norm_patchtokens"]

        # Resize the features to the expected shape
        # (B x Num_patches x Embed_dim) -> (B x Embed_dim x H / Patch_Size x W / Patch_Size)
        features = features.permute(0, 2, 1)
        features = features.reshape(
            -1,
            self.enc_embed_dim,
            height // self.final_layer_patch_size,
            width // self.final_layer_patch_size,
        ).contiguous()

        return ViTEncoderOutput(features=features)


class DINOv3ConvNextIntermediateFeatureReturner(DINOv3ConvNextEncoder, IntermediateFeatureReturner):
    "Intermediate Feature Returner for UniCeption DINOv3 ConvNext Encoder"

    def __init__(
        self,
        name: str,
        dinov3_repo_dir: str,
        data_norm_type: str = "dinov3",
        patch_size: int = 16,
        size: str = "large",
        weights: Optional[str] = None,
        pretrained_checkpoint_path: Optional[str] = None,
        indices: Optional[Union[int, List[int]]] = 1,
        norm_intermediate: bool = True,
        *args,
        **kwargs,
    ):
        """
        DINOv3 ConvNext Encoders for extracting spatial features from images.

        Args:
            name (str): Name of the encoder.
            dinov3_repo_dir (str): Path to the local directory where the DINOv3 repo was cloned. Required.
            data_norm_type (str): Image normalization type. Default: "dinov3"
            patch_size (int): Patch size for the encoder. Default: 16
            size (str): Size variant of the DINOv3 model. Options: ["tiny", "small", "base", "large"]. Default: "large"
            weights (Optional[str]): Path or URL to pretrained DINOv3 backbone checkpoint to pass to torch.hub.load. Default: None
            pretrained_checkpoint_path (str): Path to the pretrained checkpoint if using custom trained version of DINOv3. Default: None
            indices (Optional[Union[int, List[int]]], optional): Indices of the layers to return. Defaults to 1. Options:
            - int: Return the last n layers.
            - List[int]: Return the intermediate layers at the specified indices.
            norm_intermediate (bool, optional): Whether to normalize the intermediate features. Defaults to True.
        """
        # Init the base classes
        DINOv3ConvNextEncoder.__init__(
            self,
            name=name,
            data_norm_type=data_norm_type,
            patch_size=patch_size,
            size=size,
            dinov3_repo_dir=dinov3_repo_dir,
            weights=weights,
            pretrained_checkpoint_path=pretrained_checkpoint_path,
            *args,
            **kwargs,
        )
        IntermediateFeatureReturner.__init__(
            self,
            indices=indices,
            norm_intermediate=norm_intermediate,
        )

    def forward(self, encoder_input: ViTEncoderInput) -> List[ViTEncoderOutput]:
        """
        DINOv3 ConvNext Encoder Forward Pass with Intermediate Feature Return

        Args:
            encoder_input (ViTEncoderInput): Input data for the encoder. Input data must contain image normalization type and normalized image tensor.

        Returns:
            List[ViTEncoderOutput]: Output data from the encoder. Returns a list of intermediate features.
        """
        # Check image normalization type
        self._check_data_normalization_type(encoder_input.data_norm_type)

        # Check the dtype and shape of the input image
        assert isinstance(encoder_input.image, torch.Tensor), "Input must be a torch.Tensor"
        assert encoder_input.image.ndim == 4, "Input must be of shape (B, C, H, W)"
        batch_size, channels, height, width = encoder_input.image.shape
        assert channels == 3, "Input must have 3 channels"
        assert (
            height % self.patch_size == 0 and width % self.patch_size == 0
        ), f"Input shape must be divisible by patch size: {self.patch_size}"

        if self.indices is None:
            self.indices = range(len(self.model.downsample_layers))

        # Extract the intermediate features from the DINOv3 model
        intermediate_features = self.model.get_intermediate_layers(
            encoder_input.image,
            n=self.indices,
            reshape=True,
            norm=self.norm_intermediate,
        )

        # Convert the intermediate features to a list of ViTEncoderOutput
        intermediate_features = [ViTEncoderOutput(features=features) for features in intermediate_features]

        return intermediate_features


if __name__ == "__main__":
    # Set the DINOv3 repo directory and checkpoint directory
    dinov3_repo_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../local/dinov3"))
    checkpoint_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../checkpoints/encoders"))

    # Init different variants of DINOv3
    # Hash mapping for each size variant
    hash_map = {
        "small": "08c60483",
        "small+": "4057cbaa",
        "base": "73cec8be",
        "large": "8aa4cbdd",
        "huge+": "7c1da9a5",
        "7b": "a955f4ea",
    }

    for size in ["small", "small+", "base", "large", "huge+", "7b"]:
        name = f"dinov3_{size}"
        size_str = size[0] if size != "7b" else "7b"
        size_str_plus = "plus" if "+" in size else ""
        ckpt_name = f"dinov3_vit{size_str}16{size_str_plus}_pretrain_lvd1689m-{hash_map[size]}.pth"
        pretrained_checkpoint_path = os.path.join(checkpoint_dir, ckpt_name)
        dinov3_encoder = DINOv3Encoder(
            name=name,
            dinov3_repo_dir=dinov3_repo_dir,
            size=size,
            weights=pretrained_checkpoint_path,
        )

        if size == "small":
            # Run dummy forward pass
            dummy_input = ViTEncoderInput(image=torch.randn(1, 3, 224, 224), data_norm_type="dinov3")
            dummy_output = dinov3_encoder(dummy_input)

    # Init different variants of DINOv3 - ConvNext
    # Hash mapping for each ConvNext size variant
    convnext_hash_map = {
        "tiny": "21b726bb",
        "small": "296db49d",
        "base": "801f2ba9",
        "large": "61fa432d",
    }

    for size in ["tiny", "small", "base", "large"]:
        ckpt_name = f"dinov3_convnext_{size}_pretrain_lvd1689m-{convnext_hash_map[size]}.pth"
        pretrained_checkpoint_path = os.path.join(checkpoint_dir, ckpt_name)
        dinov3_convnext_encoder = DINOv3ConvNextEncoder(
            name="dino_convnext",
            dinov3_repo_dir=dinov3_repo_dir,
            size=size,
            weights=pretrained_checkpoint_path,
        )

        if size == "tiny":
            # Run dummy forward pass
            dummy_input = ViTEncoderInput(image=torch.randn(1, 3, 224, 224), data_norm_type="dinov3")
            dummy_output = dinov3_convnext_encoder(dummy_input)

    print("All DINOv3 ViT & ConvNext Encoders have been initialized successfully!")

    # Intermediate Feature Returner Tests
    print("Running Intermediate Feature Returner Tests...")

    # Run the intermediate feature returner with last-n index
    dinov3_intermediate_feature_returner = DINOv3IntermediateFeatureReturner(
        name="dinov3_small", dinov3_repo_dir=dinov3_repo_dir, size="small", indices=6
    )  # Last 6 layers
    dummy_input = ViTEncoderInput(image=torch.randn(1, 3, 224, 224), data_norm_type="dinov3")
    output = dinov3_intermediate_feature_returner(dummy_input)
    assert isinstance(output, list), "Output must be a list of intermediate features"
    assert isinstance(output[0], ViTEncoderOutput), "Output must be a list of ViTEncoderOutput"
    assert len(output) == 6, "Output must have length of intermediate features equal to the number of indices"

    # Run the intermediate feature returner with specific indices
    dinov3_intermediate_feature_returner = DINOv3IntermediateFeatureReturner(
        name="dinov3_small",
        dinov3_repo_dir=dinov3_repo_dir,
        size="small",
        indices=[0, 2, 4, 6],
    )  # Specific layers
    dummy_input = ViTEncoderInput(image=torch.randn(1, 3, 224, 224), data_norm_type="dinov3")
    output = dinov3_intermediate_feature_returner(dummy_input)
    assert isinstance(output, list), "Output must be a list of intermediate features"
    assert isinstance(output[0], ViTEncoderOutput), "Output must be a list of ViTEncoderOutput"
    assert len(output) == 4, "Output must have length of intermediate features equal to the number of indices"

    # Run the intermediate feature returner with last-n index
    dinov3_intermediate_feature_returner = DINOv3ConvNextIntermediateFeatureReturner(
        name="dinov3_convnext_small",
        dinov3_repo_dir=dinov3_repo_dir,
        size="small",
        indices=4,
    )  # Last 4 layers
    dummy_input = ViTEncoderInput(image=torch.randn(1, 3, 224, 224), data_norm_type="dinov3")
    output = dinov3_intermediate_feature_returner(dummy_input)
    assert isinstance(output, list), "Output must be a list of intermediate features"
    assert isinstance(output[0], ViTEncoderOutput), "Output must be a list of ViTEncoderOutput"
    assert len(output) == 4, "Output must have length of intermediate features equal to the number of indices"

    # Run the intermediate feature returner with specific indices
    dinov3_intermediate_feature_returner = DINOv3ConvNextIntermediateFeatureReturner(
        name="dinov3_convnext_small",
        dinov3_repo_dir=dinov3_repo_dir,
        size="small",
        indices=[0, 2],
    )  # Specific layers
    dummy_input = ViTEncoderInput(image=torch.randn(1, 3, 224, 224), data_norm_type="dinov3")
    output = dinov3_intermediate_feature_returner(dummy_input)
    assert isinstance(output, list), "Output must be a list of intermediate features"
    assert isinstance(output[0], ViTEncoderOutput), "Output must be a list of ViTEncoderOutput"
    assert len(output) == 2, "Output must have length of intermediate features equal to the number of indices"

    print("All Intermediate Feature Returner Tests have passed successfully!")
